from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional, Tuple

import mlx.core as mx
from mlx_lm.models.base import create_causal_mask
from mlx_vlm.models.qwen3_vl import LanguageModel as Qwen3VLLanguageModel
from mlx_vlm.models.qwen3_vl import Model as Qwen3VLBackbone
from mlx_vlm.models.qwen3_vl import ModelConfig, TextConfig, VisionConfig
from mlx_vlm.models.qwen3_vl import VisionModel as Qwen3VLVisionModel

from ..base import BaseModelArgs, BaseModelOutput, normalize_embeddings


def build_qwen3_vl_config(vlm_config: Dict[str, Any]) -> ModelConfig:
    base_config = dict(vlm_config)
    base_config["model_type"] = "qwen3_vl"

    config = ModelConfig.from_dict(base_config)
    if isinstance(config.text_config, dict):
        config.text_config = TextConfig.from_dict(config.text_config)
    if isinstance(config.vision_config, dict):
        config.vision_config = VisionConfig.from_dict(config.vision_config)
    return config


def last_non_padding_token(
    hidden_states: mx.array,
    attention_mask: Optional[mx.array],
) -> mx.array:
    if attention_mask is None:
        return hidden_states[:, -1]

    last_one_positions = mx.argmax(attention_mask[:, ::-1], axis=1)
    token_positions = attention_mask.shape[1] - last_one_positions - 1
    batch_positions = mx.arange(hidden_states.shape[0])
    return hidden_states[batch_positions, token_positions]


def compute_qwen3_vl_hidden_states(
    model: Qwen3VLBackbone,
    input_ids: mx.array,
    attention_mask: Optional[mx.array] = None,
    position_ids: Optional[mx.array] = None,
    inputs_embeds: Optional[mx.array] = None,
    pixel_values: Optional[mx.array] = None,
    image_grid_thw: Optional[mx.array] = None,
    video_grid_thw: Optional[mx.array] = None,
    cache=None,
) -> mx.array:
    if input_ids is None:
        raise ValueError(
            "`input_ids` is required for Qwen3-VL embedding and reranking."
        )

    visual_pos_masks = None
    deepstack_visual_embeds = None

    if inputs_embeds is None:
        input_embeddings = model.get_input_embeddings(
            input_ids=input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            video_grid_thw=video_grid_thw,
            mask=attention_mask,
        )
        inputs_embeds = input_embeddings.inputs_embeds
        visual_pos_masks = input_embeddings.visual_pos_masks
        deepstack_visual_embeds = input_embeddings.deepstack_visual_embeds

    language_model = model.language_model
    if pixel_values is not None:
        language_model._rope_deltas = None
        language_model._position_ids = None

    rope_mask = attention_mask
    model_mask = attention_mask
    if attention_mask is not None and attention_mask.ndim == 2:
        seq_length = attention_mask.shape[-1]
        causal_mask = create_causal_mask(seq_length)
        valid_tokens = attention_mask.astype(mx.bool_)
        key_mask = mx.expand_dims(valid_tokens, axis=(1, 2))
        query_mask = mx.expand_dims(valid_tokens, axis=(1, 3))
        model_mask = causal_mask[None, None, :, :] & key_mask & query_mask
    if attention_mask is not None and attention_mask.shape[-1] != input_ids.shape[-1]:
        rope_mask = None

    if position_ids is None and (rope_mask is None or rope_mask.ndim == 2):
        if (
            language_model._position_ids is not None
            and language_model._position_ids.shape[-1] >= input_ids.shape[-1]
        ):
            position_ids = language_model._position_ids[:, :, : input_ids.shape[-1]]
        else:
            position_ids, rope_deltas = language_model.get_rope_index(
                input_ids,
                image_grid_thw=image_grid_thw,
                video_grid_thw=video_grid_thw,
                attention_mask=rope_mask,
            )
            language_model._rope_deltas = rope_deltas
            language_model._position_ids = position_ids

    return language_model.model(
        input_ids,
        inputs_embeds=inputs_embeds,
        mask=model_mask,
        cache=cache,
        position_ids=position_ids,
        visual_pos_masks=visual_pos_masks,
        deepstack_visual_embeds=deepstack_visual_embeds,
    )


@dataclass
class ModelOutput(BaseModelOutput):
    logits: Optional[mx.array] = None
    scores: Optional[mx.array] = None


@dataclass
class ModelArgs(BaseModelArgs):
    text_config: Dict[str, Any]
    vision_config: Dict[str, Any]
    vlm_config: Dict[str, Any]
    model_type: str = "qwen3_vl"
    normalize: bool = True
    yes_token_id: int = 9693
    no_token_id: int = 2152

    @classmethod
    def from_dict(cls, params):
        text_config_raw = params.get("text_config", {})
        vision_config_raw = params.get("vision_config", {})

        text_config = (
            asdict(TextConfig.from_dict(text_config_raw)) if text_config_raw else {}
        )
        vision_config = (
            asdict(VisionConfig.from_dict(vision_config_raw))
            if vision_config_raw
            else {}
        )

        return cls(
            text_config=text_config,
            vision_config=vision_config,
            vlm_config=dict(params),
            model_type=params.get("model_type", "qwen3_vl"),
            normalize=params.get("normalize", True),
            yes_token_id=params.get("yes_token_id", 9693),
            no_token_id=params.get("no_token_id", 2152),
        )


class Model(Qwen3VLBackbone):
    LanguageModel = Qwen3VLLanguageModel
    VisionModel = Qwen3VLVisionModel

    def __init__(self, config: ModelArgs):
        self.args = config
        super().__init__(build_qwen3_vl_config(config.vlm_config))

    @property
    def visual(self):
        return self.vision_tower

    def get_image_features(
        self,
        pixel_values: mx.array,
        image_grid_thw: Optional[mx.array] = None,
    ) -> mx.array:
        return self.vision_tower(pixel_values, image_grid_thw)[0]

    def get_video_features(
        self,
        pixel_values: mx.array,
        video_grid_thw: Optional[mx.array] = None,
    ) -> mx.array:
        return self.vision_tower(pixel_values, video_grid_thw)[0]

    def get_binary_logits(self, pooled: mx.array) -> mx.array:
        if hasattr(self.language_model, "lm_head"):
            token_logits = self.language_model.lm_head(pooled)
        else:
            token_logits = self.language_model.model.embed_tokens.as_linear(pooled)

        return (
            token_logits[:, self.args.yes_token_id]
            - token_logits[:, self.args.no_token_id]
        )

    def __call__(
        self,
        input_ids: mx.array = None,
        attention_mask: Optional[mx.array] = None,
        position_ids: Optional[mx.array] = None,
        past_key_values=None,
        inputs_embeds: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        pixel_values_videos: Optional[mx.array] = None,
        image_grid_thw: Optional[mx.array] = None,
        video_grid_thw: Optional[mx.array] = None,
        cache=None,
        cache_position=None,
        logits_to_keep=0,
        **kwargs,
    ) -> ModelOutput:
        del kwargs, cache_position, logits_to_keep

        if pixel_values is None:
            pixel_values = pixel_values_videos
        if cache is None:
            cache = past_key_values

        hidden_states = compute_qwen3_vl_hidden_states(
            model=self,
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            video_grid_thw=video_grid_thw,
            cache=cache,
        )

        pooled = last_non_padding_token(hidden_states, attention_mask)
        text_embeds = normalize_embeddings(pooled) if self.args.normalize else pooled
        logits = self.get_binary_logits(pooled)
        scores = mx.sigmoid(logits)

        return ModelOutput(
            last_hidden_state=hidden_states,
            pooler_output=pooled,
            text_embeds=text_embeds,
            logits=logits,
            scores=scores,
        )

    def compute_scores(self, inputs: Dict[str, mx.array]) -> mx.array:
        return self(**inputs).scores

    def embed(self, inputs, processor, **kwargs) -> mx.array:
        model_inputs = processor.prepare_embedding_inputs(inputs, **kwargs)
        return self(**model_inputs).text_embeds

    def rerank(self, inputs, processor, **kwargs) -> mx.array:
        model_inputs = processor.prepare_reranker_inputs(inputs, **kwargs)
        if model_inputs is None:
            return mx.array([])
        return self(**model_inputs).scores

    def process(self, inputs, processor, **kwargs):
        if processor is None or not hasattr(processor, "prepare_model_inputs"):
            raise ValueError(
                "Qwen3-VL high-level processing requires the custom Qwen3-VL processor."
            )

        if isinstance(inputs, dict) and "documents" in inputs:
            return self.rerank(inputs, processor, **kwargs)
        return self.embed(inputs, processor, **kwargs)



from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

import mlx.core as mx
from transformers import AutoImageProcessor, AutoTokenizer
from transformers.models.qwen3_vl.processing_qwen3_vl import Qwen3VLProcessor

DEFAULT_EMBEDDING_INSTRUCTION = "Represent the user's input."
DEFAULT_RERANKING_INSTRUCTION = (
    "Given a search query, retrieve relevant candidates that answer the query."
)
DEFAULT_RERANKING_SYSTEM_PROMPT = (
    "Judge whether the Document meets the requirements based on the Query and the "
    'Instruct provided. Note that the answer can only be "yes" or "no".'
)
DEFAULT_MAX_LENGTH = 8192
IMAGE_BASE_FACTOR = 16
IMAGE_FACTOR = IMAGE_BASE_FACTOR * 2
MIN_PIXELS = 4 * IMAGE_FACTOR * IMAGE_FACTOR
MAX_PIXELS = 1800 * IMAGE_FACTOR * IMAGE_FACTOR


class _UnsupportedVideoProcessor:
    def __init__(self, merge_size: int = 2):
        self.merge_size = merge_size
        self.temporal_patch_size = 2

    def __call__(self, *args, **kwargs):
        del args, kwargs
        raise ValueError(
            "Qwen3-VL video inputs are not supported by the custom MLX processor."
        )


class Processor:
    def __init__(
        self,
        processor,
        embedding_max_length: int = DEFAULT_MAX_LENGTH,
        reranking_max_length: int = DEFAULT_MAX_LENGTH,
        min_pixels: int = MIN_PIXELS,
        max_pixels: int = MAX_PIXELS,
        default_embedding_instruction: str = DEFAULT_EMBEDDING_INSTRUCTION,
        default_reranking_instruction: str = DEFAULT_RERANKING_INSTRUCTION,
        reranking_system_prompt: str = DEFAULT_RERANKING_SYSTEM_PROMPT,
    ):
        self.processor = processor
        self.embedding_max_length = embedding_max_length
        self.reranking_max_length = reranking_max_length
        self.min_pixels = min_pixels
        self.max_pixels = max_pixels
        self.default_embedding_instruction = default_embedding_instruction
        self.default_reranking_instruction = default_reranking_instruction
        self.reranking_system_prompt = reranking_system_prompt

    @classmethod
    def from_pretrained(cls, model_path, **kwargs):
        embedding_max_length = kwargs.pop("embedding_max_length", DEFAULT_MAX_LENGTH)
        reranking_max_length = kwargs.pop("reranking_max_length", DEFAULT_MAX_LENGTH)
        min_pixels = kwargs.pop("min_pixels", MIN_PIXELS)
        max_pixels = kwargs.pop("max_pixels", MAX_PIXELS)
        default_embedding_instruction = kwargs.pop(
            "default_embedding_instruction", DEFAULT_EMBEDDING_INSTRUCTION
        )
        default_reranking_instruction = kwargs.pop(
            "default_reranking_instruction", DEFAULT_RERANKING_INSTRUCTION
        )
        reranking_system_prompt = kwargs.pop(
            "reranking_system_prompt", DEFAULT_RERANKING_SYSTEM_PROMPT
        )
        trust_remote_code = kwargs.pop("trust_remote_code", True)
        kwargs.pop("use_fast", None)

        model_dir = Path(model_path)
        is_local = model_dir.exists() and model_dir.is_dir()
        load_path = str(model_dir) if is_local else model_path
        hub_kwargs = {
            key: kwargs[key]
            for key in ["cache_dir", "force_download", "revision", "token"]
            if key in kwargs
        }

        tokenizer = AutoTokenizer.from_pretrained(
            load_path,
            trust_remote_code=trust_remote_code,
            local_files_only=is_local,
            **kwargs,
        )
        image_processor = AutoImageProcessor.from_pretrained(
            load_path,
            trust_remote_code=trust_remote_code,
            local_files_only=is_local,
            use_fast=False,
            **hub_kwargs,
        )
        processor = cls._build_processor(tokenizer, image_processor)
        return cls(
            processor=processor,
            embedding_max_length=embedding_max_length,
            reranking_max_length=reranking_max_length,
            min_pixels=min_pixels,
            max_pixels=max_pixels,
            default_embedding_instruction=default_embedding_instruction,
            default_reranking_instruction=default_reranking_instruction,
            reranking_system_prompt=reranking_system_prompt,
        )

    @staticmethod
    def _build_processor(tokenizer, image_processor):
        processor = object.__new__(Qwen3VLProcessor)
        processor.tokenizer = tokenizer
        processor.image_processor = image_processor
        processor.video_processor = _UnsupportedVideoProcessor(
            merge_size=getattr(image_processor, "merge_size", 2)
        )
        processor.chat_template = getattr(tokenizer, "chat_template", None)

        if processor.chat_template is None:
            try:
                processor.chat_template = Processor._load_chat_template(
                    tokenizer.name_or_path
                )
            except Exception:
                processor.chat_template = None

        processor.image_token = (
            getattr(tokenizer, "image_token", None) or "<|image_pad|>"
        )
        processor.video_token = (
            getattr(tokenizer, "video_token", None) or "<|video_pad|>"
        )
        processor.image_token_id = (
            getattr(tokenizer, "image_token_id", None)
            if getattr(tokenizer, "image_token_id", None) is not None
            else tokenizer.convert_tokens_to_ids(processor.image_token)
        )
        processor.video_token_id = (
            getattr(tokenizer, "video_token_id", None)
            if getattr(tokenizer, "video_token_id", None) is not None
            else tokenizer.convert_tokens_to_ids(processor.video_token)
        )
        processor.vision_start_token = (
            getattr(tokenizer, "vision_start_token", None) or "<|vision_start|>"
        )
        processor.vision_end_token = (
            getattr(tokenizer, "vision_end_token", None) or "<|vision_end|>"
        )
        processor.vision_start_token_id = (
            getattr(tokenizer, "vision_start_token_id", None)
            if getattr(tokenizer, "vision_start_token_id", None) is not None
            else tokenizer.convert_tokens_to_ids(processor.vision_start_token)
        )
        processor.vision_end_token_id = (
            getattr(tokenizer, "vision_end_token_id", None)
            if getattr(tokenizer, "vision_end_token_id", None) is not None
            else tokenizer.convert_tokens_to_ids(processor.vision_end_token)
        )

        return processor

    @staticmethod
    def _load_chat_template(model_path: str) -> str:
        path = Path(model_path)
        if path.exists() and path.is_dir():
            template_path = path / "chat_template.jinja"
        else:
            from huggingface_hub import hf_hub_download

            template_path = Path(hf_hub_download(model_path, "chat_template.jinja"))
        return template_path.read_text(encoding="utf-8")

    def __call__(self, *args, **kwargs):
        return self.processor(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.processor, name)

    @property
    def tokenizer(self):
        return self.processor.tokenizer

    @property
    def image_processor(self):
        return self.processor.image_processor

    @contextmanager
    def _padding_side(self, padding_side: str):
        original_padding_side = self.processor.tokenizer.padding_side
        self.processor.tokenizer.padding_side = padding_side
        try:
            yield
        finally:
            self.processor.tokenizer.padding_side = original_padding_side

    def _ensure_list(self, value) -> List[Any]:
        if value is None:
            return []
        if isinstance(value, list):
            return value
        return [value]

    def _append_videos(
        self,
        content: List[Dict[str, Any]],
        videos,
        fps: Optional[float] = None,
        max_frames: Optional[int] = None,
    ):
        for video in self._ensure_list(videos):
            block = {"type": "video", "video": video}
            if fps is not None:
                block["fps"] = fps
            if max_frames is not None:
                block["max_frames"] = max_frames
            content.append(block)

    def _append_images(self, content: List[Dict[str, Any]], images):
        for image in self._ensure_list(images):
            content.append(
                {
                    "type": "image",
                    "image": image,
                    "min_pixels": self.min_pixels,
                    "max_pixels": self.max_pixels,
                }
            )

    def _append_texts(self, content: List[Dict[str, Any]], texts):
        for text in self._ensure_list(texts):
            content.append({"type": "text", "text": text})

    def _format_mm_content(
        self,
        text=None,
        image=None,
        video=None,
        prefix: Optional[str] = None,
        fps: Optional[float] = None,
        max_frames: Optional[int] = None,
    ) -> List[Dict[str, Any]]:
        content: List[Dict[str, Any]] = []
        if prefix is not None:
            content.append({"type": "text", "text": prefix})

        self._append_videos(content, video, fps=fps, max_frames=max_frames)
        self._append_images(content, image)
        self._append_texts(content, text)

        if len(content) == (1 if prefix is not None else 0):
            content.append({"type": "text", "text": "NULL"})

        return content

    def format_embedding_input(self, item: Dict[str, Any]) -> List[Dict[str, Any]]:
        instruction = item.get("instruction") or self.default_embedding_instruction
        content = self._format_mm_content(
            text=item.get("text"),
            image=item.get("image"),
            video=item.get("video"),
            fps=item.get("fps"),
            max_frames=item.get("max_frames"),
        )
        return [
            {
                "role": "system",
                "content": [{"type": "text", "text": instruction}],
            },
            {"role": "user", "content": content},
        ]

    def format_reranker_inputs(
        self, payload: Dict[str, Any]
    ) -> List[List[Dict[str, Any]]]:
        instruction = payload.get("instruction", self.default_reranking_instruction)
        query = payload.get("query", {})
        documents = payload.get("documents", [])
        fps = payload.get("fps")
        max_frames = payload.get("max_frames")

        pairs = []
        for document in documents:
            contents = [
                {"type": "text", "text": f"<Instruct>: {instruction}"},
            ]
            contents.extend(
                self._format_mm_content(
                    text=query.get("text"),
                    image=query.get("image"),
                    video=query.get("video"),
                    prefix="<Query>:",
                    fps=fps,
                    max_frames=max_frames,
                )
            )
            contents.extend(
                self._format_mm_content(
                    text=document.get("text"),
                    image=document.get("image"),
                    video=document.get("video"),
                    prefix="\n<Document>:",
                    fps=fps,
                    max_frames=max_frames,
                )
            )
            pairs.append(
                [
                    {
                        "role": "system",
                        "content": [
                            {
                                "type": "text",
                                "text": self.reranking_system_prompt,
                            }
                        ],
                    },
                    {"role": "user", "content": contents},
                ]
            )
        return pairs

    def _to_mx(self, batch):
        converted = {}
        for key, value in batch.items():
            if isinstance(value, mx.array):
                converted[key] = value
            elif hasattr(value, "detach") and hasattr(value, "cpu"):
                converted[key] = mx.array(value.detach().cpu().numpy())
            elif hasattr(value, "shape"):
                converted[key] = mx.array(value)
            else:
                converted[key] = value
        return converted

    def _prepare_from_conversations(
        self,
        conversations: List[List[Dict[str, Any]]],
        padding_side: str,
        max_length: int,
        return_tensors: str = "mlx",
    ):
        with self._padding_side(padding_side):
            batch = self.processor.apply_chat_template(
                conversations,
                tokenize=True,
                return_dict=True,
                return_tensors="np",
                add_generation_prompt=True,
                padding=True,
                truncation=True,
                max_length=max_length,
            )

        if return_tensors == "np":
            return batch
        if return_tensors != "mlx":
            raise ValueError(f"Unsupported return_tensors: {return_tensors}")
        return self._to_mx(batch)

    def prepare_embedding_inputs(self, inputs, return_tensors: str = "mlx", **kwargs):
        del kwargs
        items = inputs if isinstance(inputs, list) else [inputs]
        conversations = [self.format_embedding_input(item) for item in items]
        return self._prepare_from_conversations(
            conversations,
            padding_side="right",
            max_length=self.embedding_max_length,
            return_tensors=return_tensors,
        )

    def prepare_reranker_inputs(
        self, payload: Dict[str, Any], return_tensors: str = "mlx", **kwargs
    ):
        del kwargs
        conversations = self.format_reranker_inputs(payload)
        if not conversations:
            return None
        return self._prepare_from_conversations(
            conversations,
            padding_side="left",
            max_length=self.reranking_max_length,
            return_tensors=return_tensors,
        )

    def prepare_model_inputs(self, inputs, return_tensors: str = "mlx", **kwargs):
        if isinstance(inputs, dict) and "documents" in inputs:
            return self.prepare_reranker_inputs(
                inputs, return_tensors=return_tensors, **kwargs
            )
        return self.prepare_embedding_inputs(
            inputs, return_tensors=return_tensors, **kwargs
        )



import re
from pathlib import Path

ROOT = Path(__file__).resolve().parents[2]


def test_pyproject_uses_version_module():
    pyproject = (ROOT / "pyproject.toml").read_text(encoding="utf-8")

    assert re.search(
        r"dynamic\s*=\s*\[[^\]]*\"version\"",
        pyproject,
        re.DOTALL,
    )
    assert re.search(
        r'version\s*=\s*\{\s*attr\s*=\s*"mlx_embeddings\.version\.__version__"\s*\}',
        pyproject,
    )



from .model import Model, ModelArgs, ModelOutput, TextConfig, VisionConfig
from .processor import Processor

__all__ = [
    "Model",
    "ModelArgs",
    "ModelOutput",
    "Processor",
    "TextConfig",
    "VisionConfig",
]



"""Tests for convert CLI and quantization mode (q_mode) support."""

import argparse
import importlib
from unittest.mock import MagicMock, patch

import pytest

from mlx_embeddings.convert import configure_parser

_convert_mod = importlib.import_module("mlx_embeddings.convert")


class TestConfigureParser:
    """Tests for the CLI argument parser configuration."""

    def setup_method(self):
        self.parser = configure_parser()

    def test_q_mode_default(self):
        args = self.parser.parse_args(["--hf-path", "test/model"])
        assert args.q_mode == "affine"

    def test_q_mode_valid_choices(self):
        for mode in ("affine", "mxfp4", "nvfp4", "mxfp8"):
            args = self.parser.parse_args(["--hf-path", "test/model", "--q-mode", mode])
            assert args.q_mode == mode

    def test_q_mode_invalid_choice(self):
        with pytest.raises(SystemExit):
            self.parser.parse_args(["--hf-path", "test/model", "--q-mode", "bogus"])

    def test_quantize_flag(self):
        args = self.parser.parse_args(["--hf-path", "test/model", "-q"])
        assert args.quantize is True

    def test_default_bits_and_group_size(self):
        args = self.parser.parse_args(["--hf-path", "test/model"])
        assert args.q_bits == 4
        assert args.q_group_size == 64

    def test_all_convert_args_present(self):
        args = self.parser.parse_args(
            [
                "--hf-path",
                "test/model",
                "--mlx-path",
                "/tmp/out",
                "-q",
                "--q-group-size",
                "32",
                "--q-bits",
                "8",
                "--q-mode",
                "mxfp8",
                "--dtype",
                "bfloat16",
                "--upload-repo",
                "user/repo",
                "-d",
            ]
        )
        assert args.hf_path == "test/model"
        assert args.mlx_path == "/tmp/out"
        assert args.quantize is True
        assert args.q_group_size == 32
        assert args.q_bits == 8
        assert args.q_mode == "mxfp8"
        assert args.dtype == "bfloat16"
        assert args.upload_repo == "user/repo"
        assert args.dequantize is True


class TestQuantizeModeDefaults:
    """Tests for the defaults_for_mode logic inside quantize_model."""

    def _call_defaults_for_mode(self, mode, group_size, bits):
        """Extract and call defaults_for_mode via quantize_model internals.

        We test by calling quantize_model with a mock model and inspecting
        the config output for the effective values.
        """
        import copy

        import mlx.core as mx
        import mlx.nn as nn

        from mlx_embeddings.convert import quantize_model

        linear = nn.Linear(128, 64)
        model = nn.Sequential(linear)
        config = {"model_type": "test"}

        weights, qconfig = quantize_model(model, config, group_size, bits, mode=mode)
        return (
            qconfig["quantization"]["group_size"],
            qconfig["quantization"]["bits"],
            qconfig["quantization"]["mode"],
        )

    def test_affine_defaults(self):
        gs, bits, mode = self._call_defaults_for_mode("affine", 64, 4)
        assert gs == 64
        assert bits == 4
        assert mode == "affine"

    def test_mxfp4_defaults(self):
        gs, bits, mode = self._call_defaults_for_mode("mxfp4", 0, 0)
        assert gs == 32
        assert bits == 4
        assert mode == "mxfp4"

    def test_nvfp4_defaults(self):
        gs, bits, mode = self._call_defaults_for_mode("nvfp4", 0, 0)
        assert gs == 16
        assert bits == 4
        assert mode == "nvfp4"

    def test_mxfp8_defaults(self):
        gs, bits, mode = self._call_defaults_for_mode("mxfp8", 0, 0)
        assert gs == 32
        assert bits == 8
        assert mode == "mxfp8"

    def test_user_override_group_size(self):
        gs, bits, mode = self._call_defaults_for_mode("affine", 128, 4)
        assert gs == 128
        assert bits == 4

    def test_user_override_bits(self):
        gs, bits, mode = self._call_defaults_for_mode("affine", 64, 8)
        assert gs == 64
        assert bits == 8

    def test_unsupported_mode_raises(self):
        import mlx.nn as nn

        from mlx_embeddings.convert import quantize_model

        model = nn.Sequential(nn.Linear(128, 64))
        with pytest.raises(ValueError, match="Unsupported quantization mode"):
            quantize_model(model, {}, 64, 4, mode="bogus")

    def test_config_records_mode(self):
        _, _, mode = self._call_defaults_for_mode("nvfp4", 0, 0)
        assert mode == "nvfp4"

    def test_skip_vision_recorded_in_vision_config(self):
        import mlx.nn as nn

        from mlx_embeddings.convert import quantize_model

        model = nn.Sequential(nn.Linear(128, 64))
        _, qconfig = quantize_model(
            model,
            {"vision_config": {}},
            64,
            4,
            skip_vision=True,
        )
        assert qconfig["vision_config"]["skip_vision"] is True


class TestConvertQModePassthrough:
    """Verify that convert() passes q_mode through to quantize_model."""

    @patch("glob.glob", return_value=[])
    @patch.object(_convert_mod, "save_weights")
    @patch.object(_convert_mod, "save_config")
    @patch.object(_convert_mod, "quantize_model")
    @patch.object(_convert_mod, "fetch_from_hub")
    @patch.object(_convert_mod, "get_model_path")
    def test_convert_passes_q_mode(
        self,
        mock_get_model_path,
        mock_fetch,
        mock_quantize,
        mock_save_config,
        mock_save_weights,
        mock_glob,
    ):
        from pathlib import Path

        import mlx.core as mx
        import mlx.nn as nn

        mock_get_model_path.return_value = Path("/fake/path")

        mock_model = MagicMock(spec=nn.Module)
        mock_model.parameters.return_value = {"w": mx.zeros((4, 4))}
        mock_fetch.return_value = (mock_model, {"model_type": "test"}, MagicMock())

        mock_quantize.return_value = (
            {"w": mx.zeros((4, 4))},
            {
                "model_type": "test",
                "quantization": {"group_size": 32, "bits": 4, "mode": "mxfp4"},
            },
        )

        from mlx_embeddings.convert import convert

        convert(
            hf_path="test/model",
            mlx_path="/tmp/out",
            quantize=True,
            q_mode="mxfp4",
        )

        mock_quantize.assert_called_once()
        call_kwargs = mock_quantize.call_args
        assert call_kwargs.kwargs.get("mode") == "mxfp4" or (
            len(call_kwargs.args) > 4 and call_kwargs.args[4] == "mxfp4"
        )



import argparse
import copy
import glob
import shutil
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_flatten, tree_unflatten

from .utils import (
    fetch_from_hub,
    get_model_path,
    save_config,
    save_weights,
    upload_to_hub,
)


def get_class_predicate(skip_vision: bool, q_group_size: int, weights: dict = None):
    """
    Returns a predicate function for quantization that handles vision model skipping
    and dimension compatibility checks.
    """

    def class_predicate(p, m):
        if not hasattr(m, "to_quantized"):
            return False

        if skip_vision and ("vision_model" in p or "vision_tower" in p):
            return False

        if hasattr(m, "weight"):
            if m.weight.ndim < 2 or m.weight.shape[-1] % q_group_size != 0:
                print(
                    f"Skipping quantization of {p}:"
                    f" Last dimension {m.weight.shape[-1]} is not divisible by group size {q_group_size}."
                )
                return False

        if weights:
            return p in weights

        return True

    return class_predicate


def quantize_model(
    model: nn.Module,
    config: dict,
    q_group_size: int,
    q_bits: int,
    mode: str = "affine",
    skip_vision: bool = True,
) -> Tuple:
    """
    Applies quantization to the model weights.

    Args:
        model (nn.Module): The model to be quantized.
        config (dict): Model configuration.
        q_group_size (int): Group size for quantization.
        q_bits (int): Bits per weight for quantization.
        mode (str): The quantization mode. Supported values: "affine", "mxfp4",
            "nvfp4", "mxfp8". Defaults to "affine".
        skip_vision (bool): Whether to skip vision layers. Defaults to True.

    Returns:
        Tuple: Tuple containing quantized weights and config.

    Raises:
        ValueError: If an unsupported quantization mode is specified.
    """

    def defaults_for_mode(mode: str, group_size: int, bits: int) -> Tuple[int, int]:
        mode_defaults = {
            "affine": (64, 4),
            "mxfp4": (32, 4),
            "nvfp4": (16, 4),
            "mxfp8": (32, 8),
        }
        if mode not in mode_defaults:
            raise ValueError(
                f"Unsupported quantization mode '{mode}'. "
                f"Supported modes are: {', '.join(mode_defaults.keys())}"
            )
        default_group_size, default_bits = mode_defaults[mode]
        effective_group_size = group_size if group_size else default_group_size
        effective_bits = bits if bits else default_bits
        return effective_group_size, effective_bits

    quantized_config = copy.deepcopy(config)
    effective_group_size, effective_bits = defaults_for_mode(mode, q_group_size, q_bits)

    nn.quantize(
        model,
        group_size=effective_group_size,
        bits=effective_bits,
        mode=mode,
        class_predicate=get_class_predicate(
            skip_vision=skip_vision, q_group_size=effective_group_size
        ),
    )
    quantized_config["quantization"] = {
        "group_size": effective_group_size,
        "bits": effective_bits,
        "mode": mode,
    }
    if "vision_config" in quantized_config and isinstance(
        quantized_config["vision_config"], dict
    ):
        quantized_config["vision_config"]["skip_vision"] = skip_vision
    quantized_weights = dict(tree_flatten(model.parameters()))

    return quantized_weights, quantized_config


def dequantize_model(model: nn.Module) -> nn.Module:
    """
    Dequantize the quantized linear layers in the model.

    Args:
        model (nn.Module): The model with quantized linear layers.

    Returns:
        nn.Module: The model with dequantized layers.
    """
    de_quantize_layers = []
    for name, module in model.named_modules():
        if isinstance(module, nn.QuantizedLinear):
            bias = "bias" in module
            weight = module.weight
            weight = mx.dequantize(
                weight,
                module.scales,
                module.biases,
                module.group_size,
                module.bits,
            ).astype(mx.float16)
            output_dims, input_dims = weight.shape
            linear = nn.Linear(input_dims, output_dims, bias=bias)
            linear.weight = weight
            if bias:
                linear.bias = module.bias
            de_quantize_layers.append((name, linear))
        if isinstance(module, nn.QuantizedEmbedding):
            weight = mx.dequantize(
                module.weight,
                module.scales,
                module.biases,
                module.group_size,
                module.bits,
            ).astype(mx.float16)
            num_embeddings, dims = weight.shape
            emb = nn.Embedding(num_embeddings, dims)
            emb.weight = weight
            de_quantize_layers.append((name, emb))

    if len(de_quantize_layers) > 0:
        model.update_modules(tree_unflatten(de_quantize_layers))
    return model


def convert(
    hf_path: str,
    mlx_path: str = "mlx_model",
    quantize: bool = False,
    q_group_size: int = 64,
    q_bits: int = 4,
    q_mode: str = "affine",
    dtype: str = "float16",
    upload_repo: str = None,
    revision: Optional[str] = None,
    dequantize: bool = False,
    skip_vision: bool = True,
):
    print("[INFO] Loading")
    model_path = get_model_path(hf_path, revision=revision)
    model, config, tokenizer = fetch_from_hub(
        model_path, lazy=True, path_to_repo=hf_path
    )

    weights = dict(tree_flatten(model.parameters()))
    dtype = getattr(mx, dtype)
    weights = {k: v.astype(dtype) for k, v in weights.items()}

    if quantize and dequantize:
        raise ValueError("Choose either quantize or dequantize, not both.")

    if quantize:
        print("[INFO] Quantizing")
        model.load_weights(list(weights.items()))
        weights, config = quantize_model(
            model, config, q_group_size, q_bits, mode=q_mode, skip_vision=skip_vision
        )

    if dequantize:
        print("[INFO] Dequantizing")
        model = dequantize_model(model)
        weights = dict(tree_flatten(model.parameters()))

    if isinstance(mlx_path, str):
        mlx_path = Path(mlx_path)

    del model
    save_weights(mlx_path, weights, donate_weights=True)

    for pattern in ["*.py", "*.json"]:
        files = glob.glob(str(model_path / pattern))
        for file in files:
            shutil.copy(file, mlx_path)

    tokenizer.save_pretrained(mlx_path)

    save_config(config, config_path=mlx_path / "config.json")

    if upload_repo is not None:
        upload_to_hub(mlx_path, upload_repo, hf_path, config)


def configure_parser() -> argparse.ArgumentParser:
    """
    Configures and returns the argument parser for the script.

    Returns:
        argparse.ArgumentParser: Configured argument parser.
    """
    parser = argparse.ArgumentParser(
        description="Convert Hugging Face model to MLX format"
    )

    parser.add_argument("--hf-path", type=str, help="Path to the Hugging Face model.")
    parser.add_argument(
        "--mlx-path", type=str, default="mlx_model", help="Path to save the MLX model."
    )
    parser.add_argument(
        "-q", "--quantize", help="Generate a quantized model.", action="store_true"
    )
    parser.add_argument(
        "--q-group-size", help="Group size for quantization.", type=int, default=64
    )
    parser.add_argument(
        "--q-bits", help="Bits per weight for quantization.", type=int, default=4
    )
    parser.add_argument(
        "--q-mode",
        help="The quantization mode.",
        type=str,
        default="affine",
        choices=["affine", "mxfp4", "nvfp4", "mxfp8"],
    )
    parser.add_argument(
        "--dtype",
        help="Type to save the parameters, ignored if -q is given.",
        type=str,
        choices=["float16", "bfloat16", "float32"],
        default="float16",
    )
    parser.add_argument(
        "--upload-repo",
        help="The Hugging Face repo to upload the model to.",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-d",
        "--dequantize",
        help="Dequantize a quantized model.",
        action="store_true",
        default=False,
    )
    return parser


def main():
    parser = configure_parser()
    args = parser.parse_args()
    convert(**vars(args))


if __name__ == "__main__":
    main()



"""Top-level package for MLX-Embeddings."""

import os

__author__ = """Prince Canuma"""
__email__ = "prince.gdt@gmail.com"

os.environ.setdefault("TRANSFORMERS_NO_ADVISORY_WARNINGS", "1")

from .convert import convert
from .utils import generate, load
from .version import __version__



import math
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import BaseModelArgs, BaseModelOutput, mean_pooling, normalize_embeddings
from ..llama_bidirec import LlamaBidirectionalModel
from ..llama_bidirec import ModelArgs as LlamaBidirectModelArgs
from ..siglip import SiglipVisionTransformer
from ..siglip import VisionConfig as SiglipVisionConfig


@dataclass
class VisionConfig(BaseModelArgs):
    model_type: str = "siglip_vision_model"
    hidden_size: int = 1152
    intermediate_size: int = 4304
    num_hidden_layers: int = 27
    num_attention_heads: int = 16
    num_channels: int = 3
    image_size: int = 512
    patch_size: int = 16
    layer_norm_eps: float = 1e-6
    attention_dropout: float = 0.0
    hidden_act: str = "gelu_pytorch_tanh"


@dataclass
class TextConfig(BaseModelArgs):
    model_type: str = "llama_bidirec"
    hidden_size: int = 2048
    num_hidden_layers: int = 16
    intermediate_size: int = 8192
    num_attention_heads: int = 32
    num_key_value_heads: Optional[int] = 8
    head_dim: Optional[int] = 64
    max_position_embeddings: int = 131072
    vocab_size: int = 128266
    rms_norm_eps: float = 1e-5
    rope_theta: float = 500000.0
    rope_scaling: Optional[Dict[str, Union[float, str]]] = None
    rope_traditional: bool = False
    tie_word_embeddings: bool = True
    attention_bias: bool = False
    mlp_bias: bool = False
    layer_types: Optional[List[str]] = None

    def __post_init__(self):
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads
        if self.head_dim is None:
            self.head_dim = self.hidden_size // self.num_attention_heads
        if self.layer_types is None:
            self.layer_types = ["full_attention"] * self.num_hidden_layers


@dataclass
class ModelArgs(BaseModelArgs):
    model_type: str = "llama_nemotron_vl"
    vision_config: Optional[VisionConfig] = None
    llm_config: Optional[TextConfig] = None
    downsample_ratio: float = 0.5
    force_image_size: Optional[int] = 512
    img_context_token_id: int = 128258
    select_layer: int = -1
    hidden_size: int = 2048

    def __post_init__(self):
        if isinstance(self.vision_config, dict):
            self.vision_config = VisionConfig.from_dict(self.vision_config)
        elif self.vision_config is None:
            self.vision_config = VisionConfig()
        if isinstance(self.llm_config, dict):
            self.llm_config = TextConfig.from_dict(self.llm_config)
        elif self.llm_config is None:
            self.llm_config = TextConfig()
        self.hidden_size = self.llm_config.hidden_size


def pixel_shuffle(input_tensor: mx.array, shuffle_ratio: float) -> mx.array:
    batch_size, num_patches, channels = input_tensor.shape
    patch_size = int(math.sqrt(num_patches))

    input_tensor = input_tensor.reshape(batch_size, patch_size, patch_size, -1)
    batch_size, height, width, channels = input_tensor.shape

    reshaped = input_tensor.reshape(
        batch_size, height, int(width * shuffle_ratio), int(channels / shuffle_ratio)
    )
    reshaped = reshaped.transpose(0, 2, 1, 3)
    reshaped = reshaped.reshape(
        batch_size,
        int(height * shuffle_ratio),
        int(width * shuffle_ratio),
        int(channels / (shuffle_ratio**2)),
    )
    reshaped = reshaped.transpose(0, 2, 1, 3)

    return reshaped.reshape(batch_size, -1, reshaped.shape[-1])


class VisionEncoder(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        siglip_config = SiglipVisionConfig(
            hidden_size=config.hidden_size,
            intermediate_size=config.intermediate_size,
            num_hidden_layers=config.num_hidden_layers,
            num_attention_heads=config.num_attention_heads,
            num_channels=config.num_channels,
            image_size=config.image_size,
            patch_size=config.patch_size,
            vision_use_head=True,
        )
        self.vision_model = SiglipVisionTransformer(siglip_config)

    def __call__(self, pixel_values: mx.array) -> mx.array:
        x, _ = self.vision_model(pixel_values)
        return x


class Model(nn.Module):
    def __init__(self, config: ModelArgs):
        super().__init__()
        self.config = config
        self.model_type = config.model_type

        self.vision_model = VisionEncoder(config.vision_config)

        llm_args = LlamaBidirectModelArgs.from_dict(
            {
                "model_type": "llama_bidirec",
                "hidden_size": config.llm_config.hidden_size,
                "num_hidden_layers": config.llm_config.num_hidden_layers,
                "intermediate_size": config.llm_config.intermediate_size,
                "num_attention_heads": config.llm_config.num_attention_heads,
                "num_key_value_heads": config.llm_config.num_key_value_heads,
                "head_dim": config.llm_config.head_dim,
                "max_position_embeddings": config.llm_config.max_position_embeddings,
                "vocab_size": config.llm_config.vocab_size,
                "rms_norm_eps": config.llm_config.rms_norm_eps,
                "rope_theta": config.llm_config.rope_theta,
                "rope_scaling": config.llm_config.rope_scaling,
                "rope_traditional": config.llm_config.rope_traditional,
                "tie_word_embeddings": config.llm_config.tie_word_embeddings,
                "attention_bias": config.llm_config.attention_bias,
                "mlp_bias": config.llm_config.mlp_bias,
                "layer_types": config.llm_config.layer_types,
            }
        )
        self.language_model = LlamaBidirectionalModel(llm_args)

        vit_hidden = config.vision_config.hidden_size
        ds = config.downsample_ratio
        mlp_input_size = int(vit_hidden * (1 / ds) ** 2)
        llm_hidden = config.llm_config.hidden_size

        self.mlp1 = [
            nn.LayerNorm(mlp_input_size),
            nn.Linear(mlp_input_size, llm_hidden),
            nn.GELU(),
            nn.Linear(llm_hidden, llm_hidden),
        ]

        self.downsample_ratio = ds
        self.img_context_token_id = config.img_context_token_id

    def get_extended_attention_mask(self, attention_mask):
        if attention_mask.ndim == 2:
            extended = attention_mask[:, None, None, :]
            extended = mx.repeat(extended, attention_mask.shape[-1], -2)
        elif attention_mask.ndim == 3:
            extended = attention_mask[:, None, :, :]
        else:
            raise ValueError(
                f"Wrong shape for attention_mask (shape {attention_mask.shape})"
            )
        return extended

    def extract_feature(self, pixel_values: mx.array) -> mx.array:
        vit_embeds = self.vision_model(pixel_values)
        vit_embeds = pixel_shuffle(vit_embeds, shuffle_ratio=self.downsample_ratio)

        for layer in self.mlp1:
            vit_embeds = layer(vit_embeds)

        return vit_embeds

    def _merge_input_ids_with_image_features(
        self, image_features, inputs_embeds, input_ids
    ):
        B, N, C = inputs_embeds.shape

        image_positions = input_ids == self.img_context_token_id
        image_indices = np.where(image_positions)[1].tolist()

        image_features = image_features.reshape(-1, image_features.shape[-1])
        inputs_embeds[:, image_indices, :] = image_features

        return inputs_embeds.reshape(B, N, C)

    def __call__(
        self,
        input_ids: mx.array,
        attention_mask: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
    ):
        if attention_mask is None:
            attention_mask = mx.ones(input_ids.shape)

        input_embeds = self.language_model.embed_tokens(input_ids)

        if pixel_values is not None:
            if pixel_values.ndim == 5:
                pixel_values = pixel_values[0]

            dtype = (
                self.vision_model.vision_model.embeddings.patch_embedding.weight.dtype
            )
            pixel_values = pixel_values.astype(dtype)

            vit_embeds = self.extract_feature(pixel_values)
            input_embeds = self._merge_input_ids_with_image_features(
                vit_embeds, input_embeds, input_ids
            )

        extended_mask = self.get_extended_attention_mask(attention_mask)
        extended_mask = mx.where(
            extended_mask.astype(mx.bool_),
            0.0,
            -mx.inf,
        )
        extended_mask = extended_mask.astype(
            self.language_model.embed_tokens.weight.dtype
        )

        out = self.language_model(
            input_ids, extended_mask, input_embeddings=input_embeds
        )

        text_embeds = mean_pooling(out, attention_mask)
        text_embeds = normalize_embeddings(text_embeds)

        return BaseModelOutput(
            last_hidden_state=out,
            text_embeds=text_embeds,
            pooler_output=None,
        )

    def sanitize(self, weights):
        sanitized = {}
        for k, v in weights.items():
            if "rotary_emb.inv_freq" in k:
                continue
            if "lm_head" in k:
                continue
            if "head.attention.in_proj_weight" in k:
                sanitized[k.replace("in_proj_weight", "in_proj.weight")] = v
                continue
            if "head.attention.in_proj_bias" in k:
                sanitized[k.replace("in_proj_bias", "in_proj.bias")] = v
                continue

            new_key = k

            if k.startswith("language_model.model."):
                new_key = k.replace("language_model.model.", "language_model.")

            if "patch_embedding.weight" in k and v.ndim == 4:
                # HF: (out, in, kH, kW) -> MLX: (out, kH, kW, in)
                # Skip if already in MLX format (last dim is smallest = in_channels)
                if v.shape[1] < v.shape[2]:
                    v = v.transpose(0, 2, 3, 1)

            sanitized[new_key] = v
        return sanitized

    @property
    def layers(self):
        return self.language_model.layers



import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional

import mlx.core as mx
import mlx.nn as nn
from huggingface_hub import snapshot_download
from mlx_lm.tuner.lora import LoRALinear
from mlx_vlm.models.base import create_attention_mask
from mlx_vlm.models.idefics3 import LanguageModel
from mlx_vlm.models.idefics3 import Model as VLModel
from mlx_vlm.models.idefics3 import ModelConfig as VLModelConfig
from mlx_vlm.models.idefics3 import TextConfig, VisionConfig, VisionModel
from mlx_vlm.trainer.utils import get_module_by_name, set_module_by_name


def apply_lora_adapters(model, adapter_config, adapter_weights):
    # Extract LoRA configuration
    r = adapter_config.get("r", 32)
    lora_alpha = adapter_config.get("lora_alpha", 32)
    lora_dropout = adapter_config.get("lora_dropout", 0.1)
    # Calculate scale factor
    scale = lora_alpha / r

    target_modules = set()
    for key in adapter_weights.keys():
        if key.endswith(".lora_a.weight"):
            base_name = key.replace(".lora_a", "")
            assert f"{base_name}.lora_b" in adapter_weights
            target_modules.add(base_name)
        elif key.endswith(".lora_b"):
            base_name = key.replace(".lora_b", "")
            assert f"{base_name}.lora_a" in adapter_weights
            target_modules.add(base_name)

    for name in target_modules:
        # Find the module in the model by name
        module = get_module_by_name(model, name)
        # Replace with LoRALinear
        lora_module = LoRALinear.from_base(
            linear=module,
            r=r,
            dropout=lora_dropout,
            scale=scale,
        )

        set_module_by_name(model, name, lora_module)
    return model, target_modules


@dataclass
class ModelArgs(VLModelConfig):
    embedding_dim: int = 128
    mask_non_image_embeddings: bool = False


class Model(VLModel):
    """
    ColIdefics3 model for ColVision.
    """

    def __init__(self, config: ModelArgs):
        super().__init__(config)
        assert (
            config.mask_non_image_embeddings is False
        ), "mask_non_image_embeddings is not implemeted yet in ColIdefics3."

        self.embedding_dim = config.embedding_dim
        self.linear = nn.Linear(self.config.text_config.hidden_size, self.embedding_dim)
        self.mask_non_image_embeddings = config.mask_non_image_embeddings

        # ColVision models don't use the language model head.
        self.language_model.lm_head = None

    def __call__(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        assert (
            input_ids is not None or pixel_values is not None
        ), "Either input_ids or pixel_values must be provided."
        inputs_embeds = self.get_input_embeddings(input_ids, pixel_values)

        last_hidden_state = self.call_lm_without_head(inputs_embeds=inputs_embeds)
        proj = self.linear(last_hidden_state)
        # normalize with L2 norm
        proj = proj / mx.linalg.norm(proj, axis=-1, keepdims=True)
        return proj

    def call_lm_without_head(
        self,
        inputs_embeds: mx.array,
        mask: Optional[mx.array] = None,
    ):
        """
        Call the language model without the head. Used for getting the last hidden state.
        """
        lm = self.language_model
        # for passing merged input embeddings
        h = inputs_embeds.astype(lm.norm.weight.dtype)
        cache = [None] * len(lm.layers)

        if mask is None:
            mask = create_attention_mask(h, cache)

        for layer, c in zip(lm.layers, cache):
            h = layer(h, mask, c)

        last_hidden_state = lm.norm(h)
        return last_hidden_state

    @staticmethod
    def _load_base_model_and_weights(path_or_hf_repo: str):
        """
        Loads the base model config, model instance, and weights from a local path or HF repo.
        """
        path = Path(path_or_hf_repo)
        if not path.exists():
            path = Path(
                snapshot_download(
                    repo_id=path_or_hf_repo,
                    allow_patterns=[
                        "*.json",
                        "*.safetensors",
                        "*.py",
                        "tokenizer.model",
                        "*.tiktoken",
                    ],
                )
            )

        # Load the model configuration
        with open(path / "config.json", "r") as f:
            config = json.load(f)

        # Convert the config to ModelArgs and Vision/Text Configs
        model_config = ModelArgs.from_dict(config)
        model_config.vision_config = VisionConfig.from_dict(config["vision_config"])
        model_config.text_config = TextConfig.from_dict(config["text_config"])
        model = Model(model_config)

        # Load the weights
        weight_files = list(path.glob("*.safetensors"))
        if not weight_files:
            raise FileNotFoundError(f"No safetensors found in {path}")

        weights = {}
        for wf in weight_files:
            weights.update(mx.load(str(wf)))

        weights = VLModel(model_config).sanitize(weights)
        weights = VisionModel(model_config.vision_config).sanitize(weights=weights)
        weights = LanguageModel(model_config.text_config).sanitize(weights=weights)
        return model, weights

    def sanitize_adapters(self, weights: Dict[str, mx.array]) -> Dict[str, mx.array]:
        """
        Sanitize the adapter weights to match the model's expected format.
        """
        # Remove any prefix that might be added by the adapter loading
        sanitized_weights = {
            k.replace("base_model.model.model.", ""): v for k, v in weights.items()
        }
        sanitized_weights = self.sanitize(weights=sanitized_weights)
        sanitized_weights = self.language_model.sanitize(weights=sanitized_weights)
        sanitized_weights = {
            k.replace(".lora_A.weight", ".lora_a").replace(
                ".lora_B.weight", ".lora_b"
            ): v.T
            for k, v in sanitized_weights.items()
        }
        return sanitized_weights

    @staticmethod
    def from_pretrained(path_or_hf_repo: str):
        path = Path(path_or_hf_repo)
        if not path.exists():
            path = Path(
                snapshot_download(
                    repo_id=path_or_hf_repo,
                    allow_patterns=[
                        "*.json",
                        "*.safetensors",
                        "*.py",
                        "tokenizer.model",
                        "*.tiktoken",
                    ],
                )
            )

        # Check for LoRA adapter config
        adapter_config_path = path / "adapter_config.json"
        if adapter_config_path.exists():
            with open(adapter_config_path, "r") as f:
                adapter_config = json.load(f)

            # Load base model and weights
            model, weights = Model._load_base_model_and_weights(
                adapter_config["base_model_name_or_path"]
            )

            # Load LoRA adapter weights
            adapter_weight_files = list(path.glob("*.safetensors"))
            if not adapter_weight_files:
                raise FileNotFoundError(f"No adapter safetensors found in {path}")

            adapter_weights = {}
            for awf in adapter_weight_files:
                adapter_weights.update(mx.load(str(awf)))

            adapter_weights = model.sanitize_adapters(adapter_weights)

            # Apply LoRA adapters to the model
            model, target_modules = apply_lora_adapters(
                model, adapter_config, adapter_weights
            )

            # [tm].weight -> [tm].linear.weight
            for tm in target_modules:
                weights[tm + ".linear.weight"] = weights.pop(tm + ".weight")

            weights.update(adapter_weights)
            model.load_weights(list(weights.items()), strict=True)
            return model

        # Standard model loading
        model, weights = Model._load_base_model_and_weights(str(path))
        model.load_weights(list(weights.items()), strict=True)

        return model


from typing import ClassVar, List, Optional

from PIL import Image
from transformers import BatchEncoding, Idefics3Processor

from ..colvision_processor import BaseColVisionProcessor


class Processor(BaseColVisionProcessor, Idefics3Processor):
    """
    Processor for ColIdefics3.
    Ported from PyTorch to MLX from:
    https://github.com/illuin-tech/colpali/blob/main/colpali_engine/models/idefics3/colidefics3/processing_colidefics3.py
    """

    query_prefix: ClassVar[str] = "Query: "
    query_augmentation_token: ClassVar[str] = "<end_of_utterance>"
    image_token: ClassVar[str] = "<image>"
    visual_prompt_prefix: ClassVar[str] = (
        "<|im_start|>user\n<image>Describe the image.<end_of_utterance>"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def image_token_id(self) -> int:
        return self.tokenizer.convert_tokens_to_ids(self.image_token)

    def process_images(
        self,
        images: List[Image.Image],
        context_prompts: Optional[List[str]] = None,
    ) -> BatchEncoding:
        """
        Process images for ColIdefics3.

        Args:
            images: List of PIL images.
            context_prompts: List of optional context prompts, i.e. some text description of the context of the image.
        """
        texts_doc: List[str] = []
        images = [[image.convert("RGB")] for image in images]
        if context_prompts:
            if len(images) != len(context_prompts):
                raise ValueError("Length of images and context prompts must match.")
            texts_doc = context_prompts
        else:
            texts_doc = [self.visual_prompt_prefix] * len(images)

        batch_doc = self(
            text=texts_doc,
            images=images,
            return_tensors="mlx",
            padding="longest",
        )
        # Convert all numpy arrays in batch_doc to mx.array
        for k, v in batch_doc.items():
            if hasattr(v, "dtype"):
                batch_doc[k] = mx.array(v)

        return batch_doc

    def process_queries(
        self,
        queries: List[str],
        max_length: int = 50,
        suffix: Optional[str] = None,
    ) -> BatchEncoding:
        """
        Process queries for ColIdefics3.
        """
        if suffix is None:
            suffix = self.query_augmentation_token * 10
        texts_query: List[str] = []
        for query in queries:
            query = self.query_prefix + query + suffix + "\n"
            texts_query.append(query)
        batch_query = self.tokenizer(
            text=texts_query,
            return_tensors="np",
            padding="longest",
        )
        for k, v in batch_query.items():
            if hasattr(v, "dtype"):
                batch_query[k] = mx.array(v)
        return batch_query

    def score(self, qs: List[mx.array], ps: List[mx.array], **kwargs) -> mx.array:
        """
        Compute the MaxSim score (ColBERT-like) for the given multi-vector query and passage embeddings.
        """
        return self.score_multi_vector(qs, ps, **kwargs)



#!/usr/bin/env python

"""Tests for `mlx_embeddings` package."""
import unittest
from unittest.mock import MagicMock, patch

import mlx.core as mx
import numpy as np
from mlx.utils import tree_map


class TestModels(unittest.TestCase):
    def _qwen3_vl_variant_config(self, model_type, **extra):
        config = {
            "model_type": model_type,
            "text_config": {
                "model_type": "qwen3_vl_text",
                "hidden_size": 16,
                "num_hidden_layers": 1,
                "intermediate_size": 32,
                "num_attention_heads": 2,
                "num_key_value_heads": 2,
                "rms_norm_eps": 1e-5,
                "head_dim": 8,
                "vocab_size": 32,
                "rope_theta": 1000.0,
                "max_position_embeddings": 128,
                "tie_word_embeddings": True,
                "rope_scaling": {"rope_type": "mrope", "mrope_section": [2, 1, 1]},
            },
            "vision_config": {
                "model_type": "qwen3_vl",
                "depth": 1,
                "hidden_size": 16,
                "intermediate_size": 32,
                "out_hidden_size": 16,
                "num_heads": 2,
                "patch_size": 2,
                "in_channels": 3,
                "spatial_merge_size": 2,
                "temporal_patch_size": 2,
                "num_position_embeddings": 4,
                "deepstack_visual_indexes": [],
            },
            "image_token_id": 31,
            "video_token_id": 30,
            "vision_start_token_id": 29,
            "vision_end_token_id": 28,
            "vocab_size": 32,
            "yes_token_id": 3,
            "no_token_id": 4,
        }
        config.update(extra)
        return config

    def model_test_runner(
        self,
        model,
        model_type,
        num_layers,
        last_hidden_state_is_sequence=True,
        text_embeds_is_sequence=False,
    ):
        self.assertEqual(model.config.model_type, model_type)
        if hasattr(model, "encoder"):
            self.assertEqual(len(model.encoder.layer), num_layers)
        elif hasattr(model, "model"):
            self.assertEqual(len(model.model.layers), num_layers)

        batch_size = 1
        seq_length = 5

        model.update(tree_map(lambda p: p.astype(mx.float32), model.parameters()))

        inputs = mx.array([[0, 1, 2, 3, 4]])
        outputs = model(inputs)
        self.assertEqual(outputs.last_hidden_state.dtype, mx.float32)

        # Verify last_hidden_state shape
        expected_hidden_shape = (
            (batch_size, seq_length, model.config.hidden_size)
            if last_hidden_state_is_sequence
            else (batch_size, model.config.hidden_size)
        )
        self.assertEqual(outputs.last_hidden_state.shape, expected_hidden_shape)

        output_dim = getattr(model.config, "out_features", model.config.hidden_size)

        # Verify text_embeds shape
        expected_embeds_shape = (
            (batch_size, seq_length, output_dim)
            if text_embeds_is_sequence
            else (batch_size, output_dim)
        )
        self.assertEqual(outputs.text_embeds.shape, expected_embeds_shape)
        self.assertEqual(outputs.text_embeds.dtype, mx.float32)

    def vlm_model_test_runner(self, model, model_type, num_layers):
        self.assertEqual(model.config.model_type, model_type)

        # Check layers based on model architecture
        if hasattr(model, "encoder"):
            self.assertEqual(len(model.encoder.layer), num_layers)
        # For SigLIP models, check vision and text layers separately
        elif model_type == "siglip":
            if (
                hasattr(model, "vision_model")
                and hasattr(model.vision_model, "vision_model")
                and hasattr(model.vision_model.vision_model, "encoder")
            ):
                self.assertEqual(
                    len(model.vision_model.vision_model.encoder.layers),
                    model.config.vision_config.num_hidden_layers,
                )
            if hasattr(model, "text_model") and hasattr(model.text_model, "encoder"):
                self.assertEqual(
                    len(model.text_model.encoder.layers),
                    model.config.text_config.num_hidden_layers,
                )

        batch_size = 1
        seq_length = 5

        # Convert model parameters to float32 for testing
        model.update(tree_map(lambda p: p.astype(mx.float32), model.parameters()))

        # Test text-only input if supported
        if hasattr(model, "get_text_features") or not hasattr(model, "vision_config"):
            text_inputs = mx.array([[0, 1, 2, 3, 4]])
            attention_mask = mx.ones((batch_size, seq_length))

            if hasattr(model, "get_text_features"):
                text_outputs = model.get_text_features(text_inputs, attention_mask)
                self.assertIsNotNone(text_outputs)
                self.assertEqual(text_outputs.dtype, mx.float32)
            else:
                text_outputs = model(text_inputs)
                self.assertEqual(
                    text_outputs.last_hidden_state.shape,
                    (batch_size, seq_length, model.config.hidden_size),
                )
                self.assertEqual(text_outputs.last_hidden_state.dtype, mx.float32)

        # Test image-only input if supported
        if hasattr(model, "vision_config"):
            # Get image size from vision config
            image_size = model.vision_config.image_size
            # Create dummy image tensor [batch_size, height, width, channels]
            image_inputs = mx.random.normal((batch_size, 3, image_size, image_size))

            if hasattr(model, "get_image_features"):
                image_outputs = model.get_image_features(image_inputs)
                self.assertIsNotNone(image_outputs)
                self.assertEqual(image_outputs.dtype, mx.float32)
            elif hasattr(model, "encode_image"):
                image_outputs = model.encode_image(image_inputs)
                self.assertIsNotNone(image_outputs)
                self.assertEqual(image_outputs.dtype, mx.float32)

        # Test multimodal input if model supports both text and image
        text_inputs = mx.array([[0, 1, 2, 3, 4]])
        attention_mask = mx.ones((batch_size, seq_length))
        image_size = model.config.vision_config.image_size
        image_inputs = mx.random.normal((batch_size, 3, image_size, image_size))

        # Only try this if the model has a method that accepts both inputs
        multimodal_outputs = model(
            input_ids=text_inputs,
            attention_mask=attention_mask,
            pixel_values=image_inputs,
        )

        self.assertEqual(
            multimodal_outputs.text_model_output[0].shape,
            (batch_size, seq_length, model.config.text_config.hidden_size),
        )
        self.assertEqual(
            multimodal_outputs.vision_model_output[0].shape,
            (batch_size, image_size * 2, model.config.vision_config.hidden_size),
        )
        self.assertEqual(
            multimodal_outputs.text_embeds.shape,
            (batch_size, model.config.text_config.hidden_size),
        )
        self.assertEqual(
            multimodal_outputs.image_embeds.shape,
            (batch_size, model.config.vision_config.hidden_size),
        )
        self.assertEqual(multimodal_outputs.logits_per_image.shape, (batch_size, 1))
        self.assertEqual(multimodal_outputs.logits_per_text.shape, (batch_size, 1))

    def test_xlm_roberta_model(self):
        from mlx_embeddings.models import xlm_roberta

        config = xlm_roberta.ModelArgs(
            model_type="xlm-roberta",
            hidden_size=768,
            num_hidden_layers=12,
            intermediate_size=3072,
            num_attention_heads=12,
            max_position_embeddings=512,
            vocab_size=250002,
        )
        model = xlm_roberta.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
        )

    def test_bert_model(self):
        from mlx_embeddings.models import bert

        config = bert.ModelArgs(
            model_type="bert",
            hidden_size=384,
            num_hidden_layers=6,
            intermediate_size=1536,
            num_attention_heads=12,
            max_position_embeddings=512,
            vocab_size=30522,
        )
        model = bert.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
        )

    def test_lfm2_model(self):
        from mlx_embeddings.models import lfm2

        config = lfm2.ModelArgs(
            model_type="lfm2",
            hidden_size=1024,
            num_hidden_layers=16,
            num_attention_heads=16,
            num_key_value_heads=8,
            max_position_embeddings=128000,
            vocab_size=64402,
            norm_eps=1e-05,
            layer_types=[
                "conv",
                "conv",
                "full_attention",
                "conv",
                "conv",
                "full_attention",
                "conv",
                "conv",
                "full_attention",
                "conv",
                "full_attention",
                "conv",
                "full_attention",
                "conv",
                "full_attention",
                "conv",
            ],
            conv_bias=False,
            conv_L_cache=3,
            block_dim=1024,
            block_ff_dim=6656,
            block_multiple_of=256,
            block_ffn_dim_multiplier=1.0,
            block_auto_adjust_ff_dim=True,
            rope_theta=1000000.0,
            out_features=128,
        )
        model = lfm2.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
            text_embeds_is_sequence=True,
            last_hidden_state_is_sequence=True,
        )

    def test_modernbert_model_mask_token(self):
        from mlx_embeddings.models import modernbert

        config = modernbert.ModelArgs(
            architectures=["ModernBertForMaskedLM"],
            model_type="modernbert",
            hidden_size=768,
            num_hidden_layers=22,
            intermediate_size=1152,
            num_attention_heads=12,
            max_position_embeddings=8192,
            vocab_size=50368,
        )
        model = modernbert.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
            text_embeds_is_sequence=True,
            last_hidden_state_is_sequence=True,
        )

    def test_modernbert_model_embeddings(self):
        from mlx_embeddings.models import modernbert

        config = modernbert.ModelArgs(
            architectures=["ModernBertModel"],
            model_type="modernbert",
            hidden_size=768,
            num_hidden_layers=22,
            intermediate_size=1152,
            num_attention_heads=12,
            max_position_embeddings=8192,
            vocab_size=50368,
        )
        model = modernbert.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
            last_hidden_state_is_sequence=False,
            text_embeds_is_sequence=False,
        )

    def test_siglip_model(self):
        from mlx_embeddings.models import siglip

        config = siglip.ModelArgs(
            model_type="siglip",
            text_config=siglip.TextConfig(
                hidden_size=768,
                num_hidden_layers=12,
                intermediate_size=3072,
                num_attention_heads=12,
                max_position_embeddings=512,
                vocab_size=250002,
            ),
            vision_config=siglip.VisionConfig(
                hidden_size=768,
                num_hidden_layers=12,
                intermediate_size=3072,
                num_attention_heads=12,
                image_size=512,
                patch_size=16,
            ),
        )
        model = siglip.Model(config)

        self.vlm_model_test_runner(
            model,
            config.model_type,
            config.text_config.num_hidden_layers,
        )

    def test_siglip2_model(self):
        """Test SigLIP2 with new parameters including num_patches for dynamic resolution."""
        from mlx_embeddings.models import siglip

        # Test SigLIP2 with num_patches specified (new SigLIP2 feature)
        config = siglip.ModelArgs(
            model_type="siglip",
            text_config=siglip.TextConfig(
                hidden_size=768,
                num_hidden_layers=12,
                intermediate_size=3072,
                num_attention_heads=12,
                max_position_embeddings=64,
                vocab_size=32000,
            ),
            vision_config=siglip.VisionConfig(
                hidden_size=768,
                num_hidden_layers=12,
                intermediate_size=3072,
                num_attention_heads=12,
                image_size=512,  # Same as original SigLIP test
                patch_size=16,
                num_patches=1024,  # SigLIP2 feature: (512//16)**2 = 1024
                max_num_patches=1024,  # SigLIP2 naflex feature
            ),
        )
        model = siglip.Model(config)

        # Test basic functionality
        self.vlm_model_test_runner(
            model,
            config.model_type,
            config.text_config.num_hidden_layers,
        )

        # Test SigLIP2-specific features
        import mlx.core as mx

        batch_size = 2
        image_size = config.vision_config.image_size  # Use the config's image_size
        seq_len = 64

        # Test with pixel_attention_mask and spatial_shapes (SigLIP2 naflex features)
        pixel_values = mx.random.normal((batch_size, image_size, image_size, 3))
        input_ids = mx.array([[1, 2, 3, 4, 5] + [0] * (seq_len - 5)] * batch_size)
        attention_mask = mx.ones((batch_size, seq_len))

        # SigLIP2 specific parameters
        pixel_attention_mask = mx.ones((batch_size, image_size, image_size))
        spatial_shapes = mx.array([[image_size, image_size]] * batch_size)

        # Test forward pass with SigLIP2 parameters
        outputs = model(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            pixel_attention_mask=pixel_attention_mask,
            spatial_shapes=spatial_shapes,
        )

        # Verify outputs have expected shapes
        self.assertIsNotNone(outputs.logits_per_image)
        self.assertIsNotNone(outputs.logits_per_text)
        self.assertEqual(outputs.logits_per_image.shape, (batch_size, batch_size))
        self.assertEqual(outputs.logits_per_text.shape, (batch_size, batch_size))

    def test_llama_nemotron_vl_text_only(self):
        from mlx_embeddings.models.llama_nemotron_vl.model import (
            Model as NemotronVLModel,
        )
        from mlx_embeddings.models.llama_nemotron_vl.model import (
            ModelArgs as NemotronVLModelArgs,
        )
        from mlx_embeddings.models.llama_nemotron_vl.model import (
            TextConfig,
            VisionConfig,
        )

        config = NemotronVLModelArgs(
            model_type="llama_nemotron_vl",
            hidden_size=64,
            vision_config=VisionConfig(
                hidden_size=64,
                intermediate_size=128,
                num_hidden_layers=2,
                num_attention_heads=2,
                image_size=32,
                patch_size=16,
            ),
            llm_config=TextConfig(
                hidden_size=64,
                num_hidden_layers=2,
                intermediate_size=128,
                num_attention_heads=2,
                num_key_value_heads=2,
                head_dim=32,
                max_position_embeddings=128,
                vocab_size=1000,
            ),
            downsample_ratio=0.5,
            img_context_token_id=999,
        )
        model = NemotronVLModel(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.llm_config.num_hidden_layers,
        )

    def test_llama_bidirec_model(self):
        from mlx_embeddings.models import llama_bidirec

        config = llama_bidirec.ModelArgs(
            model_type="llama_bidirec",
            hidden_size=64,
            num_hidden_layers=2,
            intermediate_size=128,
            num_attention_heads=2,
            num_key_value_heads=2,
            head_dim=32,
            max_position_embeddings=128,
            vocab_size=1000,
            rms_norm_eps=1e-5,
            rope_theta=500000.0,
        )
        model = llama_bidirec.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
        )

    def test_qwen3_model(self):
        from mlx_embeddings.models import qwen3

        config = qwen3.ModelArgs(
            model_type="qwen3",
            hidden_size=1024,
            num_hidden_layers=28,
            intermediate_size=3072,
            num_attention_heads=16,
            num_key_value_heads=8,
            head_dim=128,
            max_position_embeddings=32768,
            vocab_size=151669,
            rope_theta=1000000,
        )
        model = qwen3.Model(config)

        self.model_test_runner(
            model,
            config.model_type,
            config.num_hidden_layers,
        )

    def test_qwen3_vl_model(self):
        from mlx_embeddings.models import qwen3_vl

        config = qwen3_vl.ModelArgs.from_dict(self._qwen3_vl_variant_config("qwen3_vl"))
        model = qwen3_vl.Model(config)
        model.update(tree_map(lambda p: p.astype(mx.float32), model.parameters()))

        attention_mask = mx.ones((1, 4), dtype=mx.int32)
        outputs = model(
            input_ids=mx.array([[1, 2, 3, 4]], dtype=mx.int32),
            attention_mask=attention_mask,
        )

        self.assertEqual(outputs.last_hidden_state.shape, (1, 4, 16))
        self.assertEqual(outputs.text_embeds.shape, (1, 16))
        self.assertEqual(outputs.logits.shape, (1,))
        self.assertEqual(outputs.scores.shape, (1,))
        self.assertTrue(
            mx.allclose(mx.linalg.norm(outputs.text_embeds, axis=-1), mx.ones((1,)))
        )

    def test_qwen3_vl_model_multimodal(self):
        from mlx_embeddings.models import qwen3_vl

        config = qwen3_vl.ModelArgs.from_dict(self._qwen3_vl_variant_config("qwen3_vl"))
        model = qwen3_vl.Model(config)
        model.update(tree_map(lambda p: p.astype(mx.float32), model.parameters()))

        pixel_values = mx.random.normal((4, 3, 2, 2, 2))
        outputs = model(
            input_ids=mx.array([[1, 31, 2]], dtype=mx.int32),
            attention_mask=mx.ones((1, 3), dtype=mx.int32),
            pixel_values=pixel_values,
            image_grid_thw=mx.array([[1, 2, 2]], dtype=mx.int32),
        )

        self.assertEqual(outputs.last_hidden_state.shape, (1, 3, 16))
        self.assertEqual(outputs.text_embeds.shape, (1, 16))
        self.assertTrue(mx.all(outputs.scores >= 0.0).item())
        self.assertTrue(mx.all(outputs.scores <= 1.0).item())

    def test_qwen3_vl_processor_formats_embedding_and_reranker_inputs(self):
        from mlx_embeddings.models import qwen3_vl

        hf_processor = MagicMock()
        hf_processor.tokenizer.padding_side = "right"
        hf_processor.image_processor = MagicMock()
        hf_processor.apply_chat_template.return_value = {
            "input_ids": np.array([[1, 2, 3]], dtype=np.int32),
            "attention_mask": np.array([[1, 1, 1]], dtype=np.int32),
        }

        processor = qwen3_vl.Processor(hf_processor)

        embedding_inputs = processor.prepare_embedding_inputs(
            {"text": "hello", "instruction": "Represent this"}
        )
        self.assertEqual(embedding_inputs["input_ids"].shape, (1, 3))
        self.assertEqual(hf_processor.tokenizer.padding_side, "right")
        embedding_conversation = hf_processor.apply_chat_template.call_args.args[0][0]
        self.assertEqual(
            embedding_conversation[0]["content"][0]["text"], "Represent this"
        )
        self.assertEqual(
            embedding_conversation[1]["content"][-1], {"type": "text", "text": "hello"}
        )

        reranker_inputs = processor.prepare_reranker_inputs(
            {
                "instruction": "Rank candidates",
                "query": {"text": "query"},
                "documents": [{"text": "doc"}],
            }
        )
        self.assertEqual(reranker_inputs["attention_mask"].shape, (1, 3))
        self.assertEqual(hf_processor.tokenizer.padding_side, "right")
        reranker_conversation = hf_processor.apply_chat_template.call_args.args[0][0]
        self.assertEqual(
            reranker_conversation[0]["content"][0]["text"],
            processor.reranking_system_prompt,
        )
        self.assertEqual(
            reranker_conversation[1]["content"][0]["text"],
            "<Instruct>: Rank candidates",
        )
        self.assertIn(
            {"type": "text", "text": "<Query>:"},
            reranker_conversation[1]["content"],
        )
        self.assertIn(
            {"type": "text", "text": "\n<Document>:"},
            reranker_conversation[1]["content"],
        )

    def test_qwen3_vl_processor_from_pretrained_uses_custom_loader(self):
        from mlx_embeddings.models import qwen3_vl

        class DummyTokenizer:
            def __init__(self):
                self.chat_template = "dummy-template"
                self.padding_side = "right"
                self.name_or_path = "dummy-model"

            def convert_tokens_to_ids(self, token):
                mapping = {
                    "<|image_pad|>": 1,
                    "<|video_pad|>": 2,
                    "<|vision_start|>": 3,
                    "<|vision_end|>": 4,
                }
                return mapping[token]

        dummy_tokenizer = DummyTokenizer()
        dummy_image_processor = MagicMock()
        dummy_image_processor.merge_size = 2

        with (
            patch.object(
                qwen3_vl.processor.AutoTokenizer,
                "from_pretrained",
                return_value=dummy_tokenizer,
            ) as mock_tokenizer,
            patch.object(
                qwen3_vl.processor.AutoImageProcessor,
                "from_pretrained",
                return_value=dummy_image_processor,
            ) as mock_image_processor,
        ):
            processor = qwen3_vl.Processor.from_pretrained("dummy-model")

        mock_tokenizer.assert_called_once()
        mock_image_processor.assert_called_once()
        self.assertIs(processor.tokenizer, dummy_tokenizer)
        self.assertIs(processor.image_processor, dummy_image_processor)
        self.assertEqual(processor.processor.chat_template, "dummy-template")
        self.assertEqual(processor.processor.video_processor.merge_size, 2)

    def test_qwen3_vl_model_process_uses_high_level_processor_paths(self):
        from mlx_embeddings.models import qwen3_vl

        class DummyProcessor:
            def prepare_embedding_inputs(self, inputs, **kwargs):
                del inputs, kwargs
                return {
                    "input_ids": mx.array([[1, 2, 3]], dtype=mx.int32),
                    "attention_mask": mx.ones((1, 3), dtype=mx.int32),
                }

            def prepare_reranker_inputs(self, inputs, **kwargs):
                del kwargs
                return {
                    "input_ids": mx.array([[1, 2, 3]], dtype=mx.int32),
                    "attention_mask": mx.ones((1, 3), dtype=mx.int32),
                }

            def prepare_model_inputs(self, inputs, **kwargs):
                if isinstance(inputs, dict) and "documents" in inputs:
                    return self.prepare_reranker_inputs(inputs, **kwargs)
                return self.prepare_embedding_inputs(inputs, **kwargs)

        config = qwen3_vl.ModelArgs.from_dict(self._qwen3_vl_variant_config("qwen3_vl"))
        model = qwen3_vl.Model(config)
        model.update(tree_map(lambda p: p.astype(mx.float32), model.parameters()))

        processor = DummyProcessor()

        embeddings = model.process([{"text": "hello"}], processor=processor)
        self.assertEqual(embeddings.shape, (1, 16))

        scores = model.process(
            {
                "query": {"text": "query"},
                "documents": [{"text": "doc 1"}],
            },
            processor=processor,
        )
        self.assertEqual(scores.shape, (1,))
        self.assertTrue(mx.all(scores >= 0.0).item())
        self.assertTrue(mx.all(scores <= 1.0).item())


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import mlx.core as mx
import mlx.nn as nn
from mlx_lm.models.llama import TransformerBlock

from .base import BaseModelArgs, BaseModelOutput, mean_pooling, normalize_embeddings


@dataclass
class ModelArgs(BaseModelArgs):
    model_type: str = "llama_bidirec"
    hidden_size: int = 4096
    num_hidden_layers: int = 32
    intermediate_size: int = 14336
    num_attention_heads: int = 32
    num_key_value_heads: Optional[int] = None
    head_dim: Optional[int] = None
    max_position_embeddings: int = 131072
    vocab_size: int = 128256
    rms_norm_eps: float = 1e-5
    rope_theta: float = 500000.0
    rope_scaling: Optional[Dict[str, Union[float, str]]] = None
    rope_traditional: bool = False
    tie_word_embeddings: bool = True
    attention_bias: bool = False
    mlp_bias: bool = False
    layer_types: Optional[List[str]] = None

    def __post_init__(self):
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads
        if self.head_dim is None:
            self.head_dim = self.hidden_size // self.num_attention_heads
        if self.layer_types is None:
            self.layer_types = ["full_attention"] * self.num_hidden_layers


class LlamaBidirectionalModel(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.config = args
        self.vocab_size = args.vocab_size
        self.num_hidden_layers = args.num_hidden_layers
        self.embed_tokens = nn.Embedding(args.vocab_size, args.hidden_size)
        self.layers = [
            TransformerBlock(args=args) for _ in range(args.num_hidden_layers)
        ]
        self.norm = nn.RMSNorm(args.hidden_size, eps=args.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        mask: Optional[mx.array] = None,
        input_embeddings: Optional[mx.array] = None,
    ):
        if input_embeddings is not None:
            h = input_embeddings
        else:
            h = self.embed_tokens(inputs)

        for layer in self.layers:
            h = layer(h, mask, cache=None)

        return self.norm(h)


class Model(nn.Module):
    def __init__(self, config: ModelArgs):
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.model = LlamaBidirectionalModel(config)

    def get_extended_attention_mask(self, attention_mask, input_shape):
        if attention_mask.ndim == 3:
            extended_attention_mask = attention_mask[:, None, :, :]
        elif attention_mask.ndim == 2:
            extended_attention_mask = attention_mask[:, None, None, :]
            extended_attention_mask = mx.repeat(
                extended_attention_mask, attention_mask.shape[-1], -2
            )
        else:
            raise ValueError(
                f"Wrong shape for attention_mask (shape {attention_mask.shape})"
            )
        return extended_attention_mask

    def __call__(
        self,
        input_ids: mx.array,
        attention_mask: Optional[mx.array] = None,
    ):
        if attention_mask is None:
            attention_mask = mx.ones(input_ids.shape)

        extended_attention_mask = self.get_extended_attention_mask(
            attention_mask, input_ids.shape
        )
        extended_attention_mask = mx.where(
            extended_attention_mask.astype(mx.bool_),
            0.0,
            -mx.inf,
        )
        extended_attention_mask = extended_attention_mask.astype(
            self.model.embed_tokens.weight.dtype
        )

        out = self.model(input_ids, extended_attention_mask)

        text_embeds = mean_pooling(out, attention_mask)
        text_embeds = normalize_embeddings(text_embeds)

        return BaseModelOutput(
            last_hidden_state=out,
            text_embeds=text_embeds,
            pooler_output=None,
        )

    def sanitize(self, weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if "rotary_emb.inv_freq" in k:
                continue
            if "lm_head.weight" in k:
                continue

            new_key = k
            if not k.startswith("model."):
                new_key = f"model.{k}"

            sanitized_weights[new_key] = v
        return sanitized_weights

    @property
    def layers(self):
        return self.model.layers



from abc import ABC, abstractmethod
from typing import List, Optional, Union

import mlx.core as mx
from PIL import Image
from transformers import BatchEncoding, BatchFeature, ProcessorMixin


class BaseColVisionProcessor(ABC, ProcessorMixin):
    """
    Base class for visual retriever processors.
    Ported from PyTorch to MLX from:
    - https://github.com/illuin-tech/colpali/blob/main/colpali_engine/utils/processing_utils.py

    Removed the methods: get_topk_plaid, create_plaid_index, get_n_patches
    """

    @abstractmethod
    def process_images(
        self,
        images: List[Image.Image],
    ) -> Union[BatchFeature, BatchEncoding]:
        pass

    @abstractmethod
    def process_queries(
        self,
        queries: List[str],
        max_length: int = 50,
        suffix: Optional[str] = None,
    ) -> Union[BatchFeature, BatchEncoding]:
        pass

    @abstractmethod
    def score(
        self,
        qs: List[mx.array],
        ps: List[mx.array],
        **kwargs,
    ) -> mx.array:
        pass

    @staticmethod
    def score_single_vector(
        qs: List[mx.array],
        ps: List[mx.array],
    ) -> mx.array:
        """
        Compute the dot product score for the given single-vector query and passage embeddings using MLX.
        """
        if len(qs) == 0:
            raise ValueError("No queries provided")
        if len(ps) == 0:
            raise ValueError("No passages provided")

        qs_stacked = mx.stack(qs)
        ps_stacked = mx.stack(ps)

        scores = mx.einsum("bd,cd->bc", qs_stacked, ps_stacked)
        assert scores.shape[0] == len(
            qs
        ), f"Expected {len(qs)} scores, got {scores.shape[0]}"
        return scores.astype(mx.float32)

    @staticmethod
    def score_multi_vector(
        qs: Union[mx.array, List[mx.array]],
        ps: Union[mx.array, List[mx.array]],
        batch_size: int = 128,
    ) -> mx.array:
        """
        Compute the late-interaction/MaxSim score (ColBERT-like) for the given multi-vector
        query embeddings (`qs`) and passage embeddings (`ps`) using MLX.
        """
        if len(qs) == 0:
            raise ValueError("No queries provided")
        if len(ps) == 0:
            raise ValueError("No passages provided")

        def pad_to_max(arrays):
            max_len = max(a.shape[0] for a in arrays)
            emb_dim = arrays[0].shape[1]
            padded = []
            for a in arrays:
                pad_width = max_len - a.shape[0]
                if pad_width > 0:
                    pad = mx.zeros((pad_width, emb_dim), dtype=a.dtype)
                    padded.append(mx.concatenate([a, pad], axis=0))
                else:
                    padded.append(a)
            return mx.stack(padded)

        scores_list = []
        for i in range(0, len(qs), batch_size):
            qs_batch = pad_to_max(qs[i : i + batch_size])
            scores_batch = []
            for j in range(0, len(ps), batch_size):
                ps_batch = pad_to_max(ps[j : j + batch_size])
                # einsum: (b,n,d),(c,s,d)->(b,c,n,s)
                sim = mx.einsum("bnd,csd->bcns", qs_batch, ps_batch)
                maxsim = mx.max(sim, axis=3)  # max over s
                summed = mx.sum(maxsim, axis=2)  # sum over n
                scores_batch.append(summed)
            scores_batch = mx.concatenate(scores_batch, axis=1)
            scores_list.append(scores_batch)
        scores = mx.concatenate(scores_list, axis=0)
        assert scores.shape[0] == len(
            qs
        ), f"Expected {len(qs)} scores, got {scores.shape[0]}"
        return scores.astype(mx.float32)
